@@ -619,10 +619,6 @@ gdk_event_translate (GdkEvent *event,
 	}
 #endif /* G_ENABLE_DEBUG */
 
-      /* bits 13 and 14 in the "state" field are the keyboard group */
-#define KEYBOARD_GROUP_SHIFT 13
-#define KEYBOARD_GROUP_MASK ((1 << 13) | (1 << 14))
-      
       event->key.type = GDK_KEY_PRESS;
       event->key.window = window;
       event->key.time = xevent->xkey.time;
@@ -630,7 +626,11 @@ gdk_event_translate (GdkEvent *event,
       event->key.string = g_strdup (buf);
       event->key.length = charcount;
 
-      event->key.group = (xevent->xkey.state & KEYBOARD_GROUP_MASK) >> KEYBOARD_GROUP_SHIFT;
+      /* bits 13 and 14 in the "state" field are the keyboard group */
+#define KEYBOARD_GROUP_SHIFT 13
+#define KEYBOARD_GROUP_MASK ((1 << 13) | (1 << 14))
+      
+      event->key.group = _gdk_x11_get_group_for_state (xevent->xkey.state);
       
       break;
       
@@ -679,6 +679,8 @@ gdk_event_translate (GdkEvent *event,
       event->key.length = 0;
       event->key.string = NULL;
       
+      event->key.group = (xevent->xkey.state & KEYBOARD_GROUP_MASK) >> KEYBOARD_GROUP_SHIFT;
+
       break;
       
     case ButtonPress:
@@ -1260,9 +1262,8 @@ gdk_event_translate (GdkEvent *event,
         gdk_synthesize_window_state (window,
                                      0,
                                      GDK_WINDOW_STATE_ICONIFIED);
-      
-      if (_gdk_xgrab_window == window_private)
-	_gdk_xgrab_window = NULL;
+
+      _gdk_xgrab_check_unmap (window, xevent->xany.serial);
       
       break;
       
@@ -1790,25 +1791,19 @@ gdk_event_send_client_message_to_all_recurse (XEvent  *xev,
   unsigned char *data;
   Window *ret_children, ret_root, ret_parent;
   unsigned int ret_nchildren;
-  gint old_warnings = _gdk_error_warnings;
   gboolean send = FALSE;
   gboolean found = FALSE;
+  gboolean result = FALSE;
   int i;
 
   if (!wm_state_atom)
     wm_state_atom = gdk_x11_get_xatom_by_name ("WM_STATE");
 
-  _gdk_error_warnings = FALSE;
-  _gdk_error_code = 0;
-  XGetWindowProperty (gdk_display, xid, wm_state_atom, 0, 0, False, AnyPropertyType,
-		      &type, &format, &nitems, &after, &data);
-
-  if (_gdk_error_code)
-    {
-      _gdk_error_warnings = old_warnings;
-
-      return FALSE;
-    }
+  gdk_error_trap_push ();
+  
+  if (XGetWindowProperty (gdk_display, xid, wm_state_atom, 0, 0, False, AnyPropertyType,
+			  &type, &format, &nitems, &after, &data) != Success)
+    goto out;
 
   if (type)
     {
@@ -1818,14 +1813,9 @@ gdk_event_send_client_message_to_all_recurse (XEvent  *xev,
   else
     {
       /* OK, we're all set, now let's find some windows to send this to */
-      if (XQueryTree (gdk_display, xid, &ret_root, &ret_parent,
-		      &ret_children, &ret_nchildren) != True ||
-	  _gdk_error_code)
-	{
-	  _gdk_error_warnings = old_warnings;
-
-	  return FALSE;
-	}
+      if (!XQueryTree (gdk_display, xid, &ret_root, &ret_parent,
+		       &ret_children, &ret_nchildren))
+	goto out;
 
       for(i = 0; i < ret_nchildren; i++)
 	if (gdk_event_send_client_message_to_all_recurse (xev, ret_children[i], level + 1))
@@ -1840,16 +1830,18 @@ gdk_event_send_client_message_to_all_recurse (XEvent  *xev,
       gdk_send_xevent (xid, False, NoEventMask, xev);
     }
 
-  _gdk_error_warnings = old_warnings;
+  result = send || found;
 
-  return (send || found);
+ out:
+  gdk_error_trap_pop ();
+
+  return result;
 }
 
 void
 gdk_event_send_clientmessage_toall (GdkEvent *event)
 {
   XEvent sev;
-  gint old_warnings = _gdk_error_warnings;
 
   g_return_if_fail(event != NULL);
   
@@ -1861,8 +1853,6 @@ gdk_event_send_clientmessage_toall (GdkEvent *event)
   sev.xclient.message_type = gdk_x11_atom_to_xatom (event->client.message_type);
 
   gdk_event_send_client_message_to_all_recurse(&sev, _gdk_root_window, 0);
-
-  _gdk_error_warnings = old_warnings;
 }
 
 /*
@@ -2056,15 +2046,17 @@ static struct
   const char *xsettings_name;
   const char *gdk_name;
 } settings_map[] = {
-  { "Net/DoubleClickTime", "gtk-double-click-timeout" },
-  { "Net/DragThreshold", "gtk-drag-threshold" },
+  { "Net/DoubleClickTime", "gtk-double-click-time" },
+  { "Net/DndDragThreshold", "gtk-dnd-drag-threshold" },
+  { "Gtk/CanChangeAccels", "gtk-can-change-accels" },
   { "Gtk/ColorPalette", "gtk-color-palette" },
+  { "Gtk/FontName", "gtk-font-name" },
+  { "Gtk/KeyThemeName", "gtk-key-theme-name" },
   { "Gtk/ToolbarStyle", "gtk-toolbar-style" },
   { "Gtk/ToolbarIconSize", "gtk-toolbar-icon-size" },
   { "Net/CursorBlink", "gtk-cursor-blink" },
   { "Net/CursorBlinkTime", "gtk-cursor-blink-time" },
-  { "Net/ThemeName", "gtk-theme-name" },
-  { "Gtk/KeyThemeName", "gtk-key-theme-name" }
+  { "Net/ThemeName", "gtk-theme-name" }
 };
 
 static void
